try:
    from suss.tools.grep_code import GrepCodeTool
    from suss.tools.read_file import ReadFileTool
    from suss.tools.find_bugs import FindBugsTool
    from suss.tools.glob_files import GlobFilesTool
except ImportError:
    from tools.grep_code import GrepCodeTool
    from tools.read_file import ReadFileTool
    from tools.find_bugs import FindBugsTool
    from tools.glob_files import GlobFilesTool



# Standard library
import re
from pathlib import Path
from collections import namedtuple
from functools import cached_property

# Third party
from sortedcollections import OrderedSet
from tree_sitter_language_pack import get_parser

# Local
try:
    from suss.constants import PATH_EXCLUSIONS, SUPPORTED_LANGUAGES
except ImportError:
    from constants import PATH_EXCLUSIONS, SUPPORTED_LANGUAGES


#########
# HELPERS
#########


MAX_HEADER_SIZE = 10
MARGIN = 3
PADDING = 1
MIN_BLOCK_SIZE = 5
SHOW_FACTOR = 0.10
MAX_TO_SHOW = 25
MIN_TO_SHOW = 5

LineRange = namedtuple("LineRange", ["start", "end"])


class LineContext(object):
    def __init__(self, line_num: int):
        self.line_num = line_num
        self.scopes = OrderedSet()  # Scopes (line #s) that the line belongs to
        self.nodes = []  # AST nodes that contain the line
        self.header_parts = []  # Lines that make up the scope header

    def add_node(self, node):
        self.nodes.append(node)

    def add_header_part(self, start: int, end: int):
        size = end - start
        if not size:
            return

        self.header_parts.append((size, start, end))

    def add_scope(self, start: int):
        self.scopes.add(start)

    @cached_property
    def header(self) -> LineRange:
        header_parts = sorted(self.header_parts)
        if len(header_parts) > 1:
            size, start, end = header_parts[0]
            if size > MAX_HEADER_SIZE:
                end = start + MAX_HEADER_SIZE
        else:
            start = self.line_num
            end = self.line_num + 1

        return LineRange(start, end)


def should_index_file(file: Path) -> bool:
    # TODO: Use .gitignore

    if file.is_dotfile:
        return False

    if not file.extension:
        return False

    if not file.is_code_file:
        return False

    if file.is_boilerplate:
        return False

    return True


def should_index_dir(dir: Path) -> bool:
    # TODO: Use .gitignore
    return not any(re.search(pattern, str(dir)) for pattern in PATH_EXCLUSIONS)


def enumerate_files(root_dir: Path, sub_dir: Path | None = None):
    sub_dir = root_dir if not sub_dir else sub_dir

    for item in sub_dir.iterdir():
        if item.is_dir():
            if not should_index_dir(item):
                continue

            yield from enumerate_files(root_dir, item)
        elif item.is_file():
            file = File(root_dir, item)
            if not should_index_file(file):
                continue

            yield file


######
# MAIN
######


class File(object):
    def __init__(self, root_dir: Path, path: Path):
        self.root_dir = root_dir
        self.path = path

        try:
            with open(self.path, "r") as f:
                self.code = f.read()
        except UnicodeDecodeError:
            self.code = ""

        self.lines = self.code.splitlines()
        self.num_lines = len(self.lines) + 1
        self.line_contexts = [LineContext(i) for i in range(self.num_lines)]
        self.index_ast()

    def __hash__(self) -> str:
        return hash(self.abs_path)

    def __repr__(self) -> str:
        return self.rel_path

    @cached_property
    def last_lineno(self) -> int:
        return len(self.lines)

    @cached_property
    def content(self) -> str:
        content = ""
        for line_num, line in enumerate(self.lines):
            content += f"{line_num + 1} {line}\n"

        return content.strip("\n")

    @cached_property
    def rel_path(self) -> str:
        return str(self.path.relative_to(self.root_dir))

    @cached_property
    def abs_path(self) -> str:
        return str(self.path)

    @cached_property
    def name(self) -> str:
        return self.path.name

    @cached_property
    def extension(self) -> str:
        return self.path.suffix

    @cached_property
    def is_dotfile(self) -> bool:
        return self.name.startswith(".")

    @cached_property
    def is_code_file(self) -> bool:
        return self.extension in SUPPORTED_LANGUAGES

    @cached_property
    def language(self) -> str | None:
        if not self.is_code_file:
            return None

        return SUPPORTED_LANGUAGES[self.extension]

    @cached_property
    def is_boilerplate(self) -> bool:
        return any(re.search(pattern, self.rel_path) for pattern in PATH_EXCLUSIONS)

    def index_ast(self):
        if not self.is_code_file:
            return

        def recurse_tree(node):
            start, end = node.start_point, node.end_point
            start_line, end_line = start[0], end[0]

            self.line_contexts[start_line].add_node(node)
            self.line_contexts[start_line].add_header_part(start_line, end_line)

            for line_num in range(start_line, end_line + 1):
                self.line_contexts[line_num].add_scope(start_line)

            for child in node.children:
                recurse_tree(child)

        parser = get_parser(self.language)
        tree = parser.parse(bytes(self.code, "utf8"))
        recurse_tree(tree.root_node)

    # Adapted from Paul Gauthier's grep-ast
    def search(self, pattern: str) -> list[int]:
        parent_scopes = OrderedSet()
        matches = OrderedSet()
        show_lines = OrderedSet()

        for line_num, line in enumerate(self.lines):
            if re.search(pattern, line, re.IGNORECASE):  # 0 for no ignore
                matches.add(line_num)
                show_lines.add(line_num)

        if not matches:
            return []

        # Add padding
        for line in list(show_lines):
            for new_line in range(line - PADDING, line + PADDING + 1):
                if new_line >= self.num_lines:
                    continue
                if new_line < 0:
                    continue

                show_lines.add(new_line)

        def get_last_line_of_scope(i: int) -> int:
            line_context = self.line_contexts[i]
            last_line = max(node.end_point[0] for node in line_context.nodes)
            return last_line

        def get_all_children(node) -> list:
            children = [node]
            for child in node.children:
                children += get_all_children(child)

            return children

        def add_parent_scopes(i: int):
            nonlocal show_lines, parent_scopes

            if i in parent_scopes:
                return

            parent_scopes.add(i)

            if i >= len(self.line_contexts):
                return

            for line_num in self.line_contexts[i].scopes:
                header = self.line_contexts[line_num].header
                show_lines |= OrderedSet(range(header.start, header.end))
                last_line = get_last_line_of_scope(line_num)
                add_parent_scopes(last_line)

        def add_child_context(i: int):
            nonlocal show_lines

            if not self.line_contexts[i].nodes:
                return

            last_line = get_last_line_of_scope(i)
            size = last_line - i
            if size < MIN_BLOCK_SIZE:
                show_lines |= OrderedSet(range(i, last_line + 1))
                return

            children = []
            for node in self.line_contexts[i].nodes:
                children += get_all_children(node)

            children = sorted(
                children,
                key=lambda node: node.end_point[0] - node.start_point[0],
                reverse=True,
            )
            num_show_lines = len(show_lines)
            max_to_show = max(min(size * SHOW_FACTOR, MAX_TO_SHOW), MIN_TO_SHOW)

            for child in children:
                if len(show_lines) > num_show_lines + max_to_show:
                    break

                child_start = child.start_point[0]
                add_parent_scopes(child_start)

        # Add the bottom line + parent context
        bottom_line = self.num_lines - 2
        show_lines.add(bottom_line)
        add_parent_scopes(bottom_line)

        # Add parent context for matched lines
        for line_num in matches:
            add_parent_scopes(line_num)

        # Then add child context
        for line_num in matches:
            add_child_context(line_num)

        # Add the top margin lines of the file
        show_lines |= OrderedSet(range(MARGIN))

        # Close small gaps
        show_lines = OrderedSet(sorted(show_lines))
        closed_show_lines = OrderedSet(show_lines)
        for index in range(len(show_lines) - 1):
            if show_lines[index + 1] - show_lines[index] == 2:
                closed_show_lines.add(show_lines[index] + 1)
        for line_num, line in enumerate(self.lines):
            if line_num not in closed_show_lines:
                continue

            if (
                line.strip()
                and line_num < self.num_lines - 2
                and not self.lines[line_num + 1].strip()
            ):
                closed_show_lines.add(line_num + 1)
        show_lines = OrderedSet(sorted(closed_show_lines))

        return [i + 1 for i in show_lines]


class Chunk(object):
    def __init__(self, line_nums: list[int], file: File):
        # Line numbers are 1-indexed and not necessarily contiguous
        self.line_nums = line_nums
        self.file = file

    def __hash__(self) -> str:
        line_nums_str = ",".join(str(num) for num in self.line_nums)
        file_str = self.file.abs_path
        return hash(f"{file_str}::{line_nums_str}")

    def __repr__(self) -> str:
        return self.to_string()

    @cached_property
    def lines(self) -> list[str]:
        return [self.file.lines[i - 1] for i in self.line_nums]

    # TODO: Ensure string doesn't start or end with dots
    def to_string(self, line_nums: bool = True, dots: bool = True) -> str:
        output_str = ""

        use_dots = not (1 in self.line_nums)
        for line_num, line in enumerate(self.file.lines):
            line_num = line_num + 1

            if line_num not in self.line_nums:
                if use_dots and dots:
                    output_str += "⋮...\n"
                    use_dots = False

                continue

            output_str += f"{line_num} {line}\n" if line_nums else f"{line}\n"
            use_dots = True

        return output_str.strip("\n")


class Index(object):
    def __init__(self, root_dir: Path):
        self.root_dir = root_dir
        self._files = {f.rel_path: f for f in enumerate_files(self.root_dir)}

    @property
    def files(self) -> list[str]:
        return list(self._files.values())

    def get_file(self, rel_path: str) -> File:
        return self._files[rel_path]

    def search_code(self, pattern: str, exclude: File | None = None) -> list[Chunk]:
        results = OrderedSet()
        for file in self._files.values():
            if exclude and file.rel_path == exclude.rel_path:
                continue

            line_matches = file.search(pattern)
            if not line_matches:
                continue

            chunk = Chunk(line_matches, file)
            results.add(chunk)

        return list(results)



try:
    from suss.constants.languages import SUPPORTED_LANGUAGES
    from suss.constants.boilerplate import PATH_EXCLUSIONS
    from suss.constants.agent import (
        MAX_MERGE_DISTANCE,
        CODE_SEARCH_LIMIT,
        FILE_SEARCH_LIMIT,
        MAX_CONTEXT_TOKENS,
        MAX_TOOL_CALLS,
        MAX_THINKING_TOKENS,
    )
except ImportError:
    from constants.languages import SUPPORTED_LANGUAGES
    from constants.boilerplate import PATH_EXCLUSIONS
    from constants.agent import (
        MAX_MERGE_DISTANCE,
        CODE_SEARCH_LIMIT,
        FILE_SEARCH_LIMIT,
        MAX_CONTEXT_TOKENS,
        MAX_TOOL_CALLS,
        MAX_THINKING_TOKENS,
    )



# Standard library
import re
import os
import math
from collections import defaultdict

# Third party
import cohere
from saplings.abstract import Tool
from tree_sitter_language_pack import get_parser

# Local
try:
    from suss.index import Index, File, Chunk
    from suss.constants import CODE_SEARCH_LIMIT
except ImportError:
    from index import Index, File, Chunk
    from constants import CODE_SEARCH_LIMIT


#########
# HELPERS
#########


try:
    client = cohere.AsyncClient(os.environ["COHERE_API_KEY"])
except:
    client = None

MAX_CHUNKS = 10000


def tokenize_query(query: str) -> list[str]:
    tokens = re.findall(r"\w+", query)
    return tokens


# TODO: Make this less dumb
def tokenize_code(chunk: Chunk) -> list[str]:
    def extract_tokens(node, code_bytes: bytes) -> list[str]:
        tokens = []
        if node.child_count == 0:
            token = code_bytes[node.start_byte : node.end_byte].decode(
                "utf8", errors="ignore"
            )

            # Skip tokens that are only whitespace or punctuation
            if token.strip() and not re.fullmatch(r"\W+", token):
                tokens.append(token.lower())
        else:
            for child in node.children:
                tokens.extend(extract_tokens(child, code_bytes))

        return tokens

    parser = get_parser(chunk.file.language)
    code_bytes = bytes(chunk.to_string(False, False), "utf8")
    tree = parser.parse(code_bytes)
    return extract_tokens(tree.root_node, code_bytes)


def bm25_rerank(
    query: str, chunks: list[Chunk], k1: float = 1.5, b: float = 0.75
) -> list[Chunk]:
    if not chunks:
        return chunks

    tokenized_query = tokenize_query(query)
    tokenized_chunks = [tokenize_code(chunk) for chunk in chunks]
    N = len(tokenized_chunks)

    token_freq = {}
    for tokens in tokenized_chunks:
        for token in set(tokens):
            token_freq[token] = token_freq.get(token, 0) + 1

    avgdl = sum(len(tokens) for tokens in tokenized_chunks) / N

    # Compute BM25 scores
    scores = []
    for i, tokens in enumerate(tokenized_chunks):
        score = 0.0

        token_counts = {}
        for t in tokens:
            token_counts[t] = token_counts.get(t, 0) + 1

        dl = len(tokens)
        for t in tokenized_query:
            if t not in token_counts:
                continue

            n_t = token_freq.get(t, 0)
            idf = math.log((N - n_t + 0.5) / (n_t + 0.5) + 1)
            tf = token_counts[t]
            score += idf * (tf * (k1 + 1)) / (tf + k1 * (1 - b + b * (dl / avgdl)))

        scores.append((score, i))

    ranked_indices = sorted(scores, key=lambda x: x[0], reverse=True)
    ranked_chunks = [chunks[i] for _, i in ranked_indices]
    return ranked_chunks


async def neural_rerank(query: str, chunks: list[Chunk]) -> list[Chunk]:
    chunk_strs = [chunk.to_string(line_nums=False) for chunk in chunks[:MAX_CHUNKS]]
    response = await client.rerank(
        model="rerank-v3.5", query=query, documents=chunk_strs
    )
    for result in response.results:
        score = result.relevance_score
        index = result.index
        chunks[index].score = score

    ranked_chunks = sorted(chunks, key=lambda c: c.score, reverse=True)
    return ranked_chunks


def query_to_regex(query: str) -> str:
    return "|".join(map(re.escape, query.split()))


async def rerank_chunks(query: str, chunks: list[Chunk]) -> list[Chunk]:
    if os.getenv("COHERE_API_KEY", None):
        return await neural_rerank(query, chunks)

    return bm25_rerank(query, chunks)


async def filter_chunks(query: str, chunks: list[Chunk]) -> list[Chunk]:
    return chunks  # TODO: LLM-based filtering


######
# MAIN
######


class GrepCodeTool(Tool):
    def __init__(
        self, index: Index, target_file: File, update_progress: callable, **kwargs
    ):
        # Base attributes
        self.name = "search_code"
        self.description = "Searches the contents of files in the codebase using regular expressions. Returns code snippets containing exact matches."
        self.parameters = {
            "type": "object",
            "properties": {
                "intent": {
                    "type": "string",
                    "description": "Concise, one-sentence description of your intent behind the search. E.g. 'Find the definition of handle_auth', 'Track lifecycle of connection_pool', 'Understand how the parser is used'.",
                },
                "query": {
                    "type": "string",
                    "description": "A search query, passed into Python's re.match(). Should match symbols in the codebase.",
                },
            },
            "required": ["intent", "query"],
            "additionalProperties": False,
        }
        self.is_terminal = False

        # Additional attributes
        self.index = index
        self.target_file = target_file
        self.update_progress = update_progress

    def format_output(self, chunks: list[Chunk]) -> str:
        grouped_chunks = defaultdict(list)
        for chunk in chunks:
            grouped_chunks[chunk.file].append(chunk)

        formatted_chunks = []
        for file, chunks in grouped_chunks.items():
            line_nums = set()
            for chunk in chunks:
                line_nums |= set(chunk.line_nums)
            line_nums = list(line_nums)
            line_nums.sort()

            chunk = Chunk(line_nums, file)
            formatted_chunk = f"<file_path>{file.rel_path}</file_path>\n<file_content>\n{chunk.to_string()}\n</file_content>"
            formatted_chunks.append(formatted_chunk)

        formatted_chunks = "\n\n".join(formatted_chunks)
        return formatted_chunks

    async def run(self, intent: str, query: str, **kwargs) -> list[Chunk]:
        self.update_progress(intent)
        query_regex = query_to_regex(query)
        results = self.index.search_code(query_regex, exclude=self.target_file)
        results = await rerank_chunks(query, results)
        results = await filter_chunks(query, results)
        return results[:CODE_SEARCH_LIMIT]


# TODO: Implement a semantic fallback if AST-grep fails to retrieve enough chunks.
# 1. Get the code map (c-tags) for the codebase.
# 2. Use the code map and the query to generate a hypothetical code snippet.
# 3. Run an (AST-based) keyword search on the codebase using the snippet.
# 4. Rerank the results.



# Standard library
import re
from collections import defaultdict

# Third party
from saplings.abstract import Tool

# Local
try:
    from suss.index import Index, File
    from suss.constants import FILE_SEARCH_LIMIT
except ImportError:
    from index import Index, File
    from constants import FILE_SEARCH_LIMIT


#########
# HELPERS
#########


def additions(query: str, i: int, j: int) -> str:
    if i > j:
        return additions(query, j, i)

    query_chars = list(query)
    query_chars.insert(j, ".?")
    query_chars.insert(i, ".?")

    for index, char in enumerate(query_chars):
        if index in [i, j + 1]:
            continue

        query_chars[index] = re.escape(char)

    return "".join(query_chars)


def replacements(query: str, i: int, j: int) -> str:
    if i > j:
        return replacements(query, j, i)

    query_chars = list(query)
    query_chars.pop(j)
    query_chars.insert(j, ".?")
    query_chars.pop(i)
    query_chars.insert(i, ".?")

    for index, char in enumerate(query_chars):
        if index in [i, j]:
            continue

        query_chars[index] = re.escape(char)

    return "".join(query_chars)


def one_of_each(query: str, i: int, j: int) -> str:
    if i > j:
        return one_of_each(query, j, i)

    query_chars = list(query)
    query_chars.pop(j)
    query_chars.insert(j, ".?")
    query_chars.insert(i, ".?")

    for index, char in enumerate(query_chars):
        if index in [i, j + 1]:
            continue

        query_chars[index] = re.escape(char)

    return "".join(query_chars)


def build_fuzzy_regex_filters(query: str) -> list[re.Pattern]:
    filters = []
    query_len = len(query)
    for i in range(query_len):
        for j in range(i, query_len):
            if j != query_len:
                filters.append(
                    re.compile(one_of_each(query, i, j), flags=re.IGNORECASE)
                )
                filters.append(
                    re.compile(replacements(query, i, j), flags=re.IGNORECASE)
                )

            filters.append(re.compile(additions(query, i, j), flags=re.IGNORECASE))

    return filters


def get_trigrams(query: str) -> list[str]:
    query_chars = list(query)

    trigrams = []
    while len(query_chars):
        if len(query_chars) <= 3:
            trigrams.append("".join(query_chars).lower())
            query_chars.clear()
        else:
            trigrams.append("".join(query_chars[:3]).lower())
            query_chars.pop(0)

    return trigrams


######
# MAIN
######


class GlobFilesTool(Tool):
    def __init__(
        self, index: Index, target_file: File, update_progress: callable, **kwargs
    ):
        # Base attributes
        self.name = "search_files"
        self.description = "Runs a fuzzy keyword search on the file paths in the codebase. Returns file paths that may not exactly match the query, but will be roughly similar."
        self.parameters = {
            "type": "object",
            "properties": {
                "intent": {
                    "type": "string",
                    "description": "Concise, one-sentence description of your intent behind the search. E.g. 'Find files in the server directory', 'Locate the DTOs module', 'Get all files related to tooling'.",
                },
                "query": {
                    "type": "string",
                    "description": "A search query with NO whitespace. Should match the file paths you're looking for. E.g. 'worker/engine', 'dtos.py', 'utils/api.ts'.",
                },
            },
            "required": ["intent", "query"],
            "additionalProperties": False,
        }
        self.is_terminal = False

        # Additional attributes
        self.index = index
        self.target_file = target_file
        self.update_progress = update_progress

    def format_output(self, files: list[File]) -> str:
        file_paths = (file.rel_path for file in files)
        return "\n".join(file_paths)

    async def run(self, intent: str, query: str, **kwargs) -> list[File]:
        self.update_progress(intent)

        query = query.lower()
        trigrams = get_trigrams(query)
        filters = build_fuzzy_regex_filters(query)

        # Get no. of trigrams that match the query, per file path
        file_counts = defaultdict(int)
        for file in self.index.files:
            for trigram in trigrams:
                if trigram not in file.rel_path.lower():
                    continue

                file_counts[file] += 1

        # Sort files by no. of trigram matches
        file_counts = sorted(
            file_counts.items(), key=lambda kv: (kv[1], kv[0].rel_path), reverse=True
        )

        # Filter file paths by edit distance
        matches = []
        for file, _ in file_counts:
            if file.rel_path == self.target_file.rel_path:
                continue

            for fuzzy_filter in filters:
                if fuzzy_filter.search(file.rel_path):
                    matches.append(file)
                    break

            if len(matches) >= FILE_SEARCH_LIMIT:
                break

        return matches



import warnings

warnings.filterwarnings("ignore")

# Standard library
import asyncio
import argparse
import subprocess
from pathlib import Path
from dataclasses import dataclass

# Third party
from rich.rule import Rule
from rich.live import Live
from rich.text import Text
from rich.table import Table
from rich.syntax import Syntax
from rich.markdown import Markdown
from rich.console import Console, Group
from rich.progress import Progress, TextColumn, BarColumn

# Local
try:
    from suss.agent import Agent
    from suss.index import Index, File
    from suss.constants import MAX_TOOL_CALLS
except ImportError:
    from agent import Agent
    from index import Index, File
    from constants import MAX_TOOL_CALLS


#########
# HELPERS
#########


@dataclass
class Config:
    file: str | None = None  # Specific file to scan for bugs
    root_dir: str | None = None
    max_iters: int = MAX_TOOL_CALLS
    model: str = "openai/gpt-4o"


def clean_root_dir(root_dir: str | None = None) -> Path:
    if root_dir is None:
        return Path.cwd()

    return Path(root_dir)


def get_config() -> Config | None:
    parser = argparse.ArgumentParser(
        description=("Use AI to catch bugs in your code changes.")
    )
    parser.add_argument(
        "--file",
        type=str,
        required=False,
        default=None,
        help="Relative path to a specific file to scan for bugs.",
    )
    parser.add_argument(
        "--root-dir",
        type=str,
        required=False,
        default=None,
        help="Root directory of your codebase.",
    )
    parser.add_argument(
        "--max-iters",
        required=False,
        type=int,
        default=MAX_TOOL_CALLS,
        help="Max. # of iterations to run the agent per file.",
    )
    parser.add_argument(
        "--model",
        type=str,
        default="openai/gpt-4o",
        help="The LLM model to use.",
    )
    args = parser.parse_args()

    return Config(
        file=args.file,
        root_dir=clean_root_dir(args.root_dir),
        max_iters=args.max_iters,
        model=args.model,
    )


def get_changed_files(index: Index, config: Config) -> list[File]:
    try:
        if config.file:
            return [f for f in index.files if f.rel_path.lower() == config.file.lower()]

        status_output = subprocess.check_output(
            ["git", "status", "--porcelain"], text=True
        ).splitlines()

        changed_files = set()
        for line in status_output:
            if not line.strip():
                continue

            status = line[:2].strip()
            file_path = line[3:].strip()

            if status == "D":  # Skip deleted files
                continue

            changed_files.add(file_path)

        is_changed_file = lambda f: any(
            c_file.strip().lower() in f.rel_path.strip().lower()
            for c_file in changed_files
        )
        return [f for f in index.files if is_changed_file(f)]
    except subprocess.SubprocessError as e:
        return []


async def run_agent(agent: Agent, file: File, progress: Progress, update_log_panel):
    task_id = progress.add_task(file.rel_path, total=agent.max_iters + 1, message="")

    def update_progress(message: str):
        progress.advance(task_id)
        progress.update(task_id, message=message)
        update_log_panel(message, file.rel_path)

    output = await agent.run(file, update_progress=update_progress)
    progress.update(task_id, completed=agent.max_iters + 1)
    return output


def print_bug_report(files: list[File], bug_sets: list[object], padding: int = 3):
    console = Console()
    # console.print("[bold]Bug Report:[/bold]")
    for file, bugs in zip(files, bug_sets):
        for bug in bugs:
            file_path, confidence = file.rel_path, bug.confidence
            bug_description = bug.description
            start, end = bug.start, bug.end
            padded_start = max(1, start - padding)
            padded_end = min(end + padding, len(file.lines))
            code = "\n".join(file.lines[padded_start - 1 : padded_end])

            # Divider
            console.print(Rule(style="cyan"))

            # Title
            table = Table.grid(expand=True)
            table.add_column(justify="left")
            table.add_column(justify="right")
            lines = f"L{start}-{end}" if start != end else f"L{start}"
            left = f"[bold][cyan]{file_path}[/cyan][/bold] [cyan]({lines})[/cyan]"
            right = f"{confidence}%"
            if confidence <= 33:
                right = f"[red]{confidence}%[/red]"
            elif confidence <= 66:
                right = f"[yellow]{confidence}%[/yellow]"
            else:
                right = f"[green]{confidence}%[/green]"
            right += " confidence"
            table.add_row(left, right)
            console.print(table)
            console.print()

            # Code block
            console.print(
                Syntax(
                    code,
                    file.language,
                    theme="monokai",
                    line_numbers=True,
                    start_line=padded_start,
                    highlight_lines=list(range(start, end + 1)),
                )
            )
            console.print()

            # Bug description
            console.print(
                Markdown(
                    bug_description,
                    code_theme="monokai",
                    inline_code_lexer="python",
                    inline_code_theme="monokai",
                )
            )
            console.print()


######
# MAIN
######


def main():
    async def async_main():
        console = Console()
        config = get_config()
        index = Index(config.root_dir)
        agent = Agent(index, config.model, config.max_iters)
        files_to_analyze = get_changed_files(index, config)

        if not files_to_analyze:
            if not config.file:
                console.print("No code changes detected. Aborting analysis.")
            else:
                console.print(f"No such file: {config.file}. Aborting analysis.")

            return

        console.print("[bold]Analyzing files...[/bold]")
        console.print("")

        log_panel = Text("")
        file_logs = []

        progress = Progress(
            TextColumn("[progress.description]{task.description}", justify="left"),
            BarColumn(bar_width=None),
            TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
            console=console,
            expand=True,
        )

        display_group = Group(progress, Text(""), Rule(style="cyan"), log_panel)

        with Live(display_group, refresh_per_second=10, transient=False) as live:

            def update_log_panel(message: str, file_path: str):
                nonlocal file_logs
                file_logs.append(f"[cyan]{file_path}:[/cyan] {message}")
                if len(file_logs) > 5:
                    file_logs.pop(0)
                display_group.renderables[-1] = Text.from_markup("\n".join(file_logs))

            tasks = []
            for file in files_to_analyze:
                task = run_agent(agent, file, progress, update_log_panel)
                tasks.append(task)

            bug_sets = await asyncio.gather(*tasks)

        console.clear()
        print_bug_report(files_to_analyze, bug_sets)

    asyncio.run(async_main())
